"""Provides some basic mathematical utility functions."""
import numpy

def rotation_matrix_from_to(a,b):
    """Returns rotation matrix R such that norm(b)*dot(R,a)/norm(a) = b.

    Args:
        a (numpy.ndarray): A 3-vector
        b (numpy.ndarray): Another 3-vector

    Returns:
        R (numpy.ndarray): The 3x3 rotation matrix that will would rotate a parallel to b.

    """
    a1 = a/numpy.linalg.norm(a)
    b1 = b/numpy.linalg.norm(b)
    theta = numpy.arccos(numpy.dot(a1,b1))
    if theta<1e-6 or numpy.isnan(theta):
        return numpy.identity(3)
    if numpy.pi-theta<1e-6: #TODO(Eric): verify correct
        d = numpy.array([1.,0,0])
        x = numpy.cross(a1,d)
    else:
        x = numpy.cross(a1,b1)
        x /= numpy.linalg.norm(x)
    A = numpy.array([ [0,-x[2],x[1]], [x[2],0,-x[0]], [-x[1],x[0],0] ])
    R = numpy.identity(3) + numpy.sin(theta)*A + (1.-numpy.cos(theta))*numpy.dot(A,A)
    return R

def shift_indicies_in_list(alist,min_i,shift):
    """Adds 'shift' to all indices>=min_i in a bond, angle, or dihedral list.

    Note that this function assumes that every element in alist is a list of strings
    where the first string is the type associated with the bond, angle, or dihedral.

    Args:
        alist (list): The list to be shifted.
        min_i (int): All indices at least this small will be shifted.
        shift (int): Amount to shift each shifted index.

    Returns:
        blist (list): shifted list
    """
    blist = []
    for i,a in enumerate(alist):
        b = [int(x) for x in a[1:]]
        for j,bb in enumerate(b):
            if bb>=min_i:
                b[j]=bb+shift
        blist.append([alist[i][0]] + [str(x)for x in b])
    return blist

def alphabetize(to_fix):
    """Cleans up bond, angle, and dihedral lists.

    Finds unique elements of bond, angle, and dihedral lists, with special care
    taken so that reversals an element type are considered (e.g., AAAB equals BAAA,
    but only the former is stored because it has lower lexicographical ordering)
    
    Args:
        to_fix (list): The first element of each element of to_fix is the type that will
            be checked and fixed.

    Returns:
        fixed_list
    """
    temp =list(set(to_fix)) #Get unique elements
    types = []
    for t in temp:
        if t==t[::-1]:
            types.append(t) #keep palindromes
        elif t<t[::-1]:
            types.append(t) #and keep the forward or reverse that has the lower lexicographical order
        else:
            types.append(t[::-1])
    types = sorted(list(set(types)))
    for i,t in enumerate(to_fix):
        if t[::-1] in types:
            to_fix[i]=t[::-1]
    return to_fix

def remove_items_not_containing(alist,key):
    """Removes all elements of a bond, angle, or dihedral list that don't contain 'key'.

    Note that this function assumes that every element in alist is a list of strings
    where the first string is the type associated with the bond, angle, or dihedral.

    Args:
        alist (list): The list to have elements removed from it
        key (str): Key to search for

    Returns:
        blist (list): shifted list
    """
    blist = []
    for a in alist:
        if key in a:
            blist.append(a)
    return blist

def remove_items_containing(alist,key):
    """Removes all elements of a bond, angle, or dihedral list containing key.

    Note that this function assumes that every element in alist is a list of strings
    where the first string is the type associated with the bond, angle, or dihedral.

    Args:
        alist (list): The list to have elements removed from it
        key (str): Key to search for

    Returns:
        blist (list): shifted list
    """
    blist = []
    for a in alist:
        if key not in a:
            blist.append(a)
    return blist


